#include <vector>
#include "finite_field.h"
#include "matrix.h"

using namespace std;

const int Poly_9[]={1,0};
FiniteField F9(3,2,Poly_9);

const int Poly_16[]={1,1,1,1};
FiniteField F16(2,4,Poly_16);

FiniteField F(F16);

int main()
{
    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }
    cout << "Körper angelegt" << endl;

    unsigned int matrix_size=2;
    unsigned int distance=2;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (hermitian)" << endl << endl;

    int q = sqrt(F.number_of_field_elements); // F = F_{q^2}

    vector<FieldElement<&F> > field_conjugated(F.number_of_field_elements);
    vector<FieldElement<&F> > f_q;
    for (int k=0; k<F.number_of_field_elements; ++k)
    {
        field_conjugated[k] = pow(field[k],q);
        if (field[k]==field_conjugated[k])
            f_q.push_back(field[k]);
    }

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_triangles = pow(F.number_of_field_elements,(matrix_size*(matrix_size-1))/2);
    int number_of_diagonals = pow(q,matrix_size);
    int number_of_herm_matrices = number_of_triangles * number_of_diagonals;
    vector<Matrix<FieldElement<&F> > > Mat_herm(number_of_herm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size-1))/2, 0);
        vector<vector<int> > Mat_index(number_of_triangles, M_index);

        for (int k=1; k<number_of_triangles; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size-1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_triangles; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
               // Mat_herm[k][r][r]= 0; entry is already zero
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_herm[k][r][c]= field[Mat_index[k][j]];
                    Mat_herm[k][c][r]= field_conjugated[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }
    {
        vector<int> M_index(matrix_size, 0);
        vector<vector<int> > Mat_index(number_of_diagonals, M_index);

        for (int k=1; k<number_of_diagonals; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< matrix_size-1; ++d)
            {
                if (Mat_index[k][d] == q)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }


        for (int k=1; k<number_of_diagonals; ++k)
        {
            for (int j=0; j<number_of_triangles; ++j)
            {
                Mat_herm[j+k*number_of_triangles] = Mat_herm[j];
                for (int i=0; i<matrix_size; ++i)
                    Mat_herm[j+k*number_of_triangles][i][i] = f_q[Mat_index[k][i]];
            }
        }
    }

    cout << "hermitesche Matrizen angelegt" << endl;


int A[] = {0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15, 16, 24, 20, 28, 18, 26, 22, 30, 17, 25, 21, 29, 19
, 27, 23, 31, 32, 40, 36, 44, 34, 42, 38, 46, 33, 41, 37, 45, 35, 43, 39, 47, 48, 56, 52, 60, 50, 58
, 54, 62, 49, 57, 53, 61, 51, 59, 55, 63, 64, 72, 68, 76, 66, 74, 70, 78, 65, 73, 69, 77, 67, 75, 71
, 79, 80, 88, 84, 92, 82, 90, 86, 94, 81, 89, 85, 93, 83, 91, 87, 95, 96, 104, 100, 108, 98, 106
, 102, 110, 97, 105, 101, 109, 99, 107, 103, 111, 112, 120, 116, 124, 114, 122, 118, 126, 113
, 121, 117, 125, 115, 123, 119, 127, 128, 136, 132, 140, 130, 138, 134, 142, 129, 137, 133
, 141, 131, 139, 135, 143, 144, 152, 148, 156, 146, 154, 150, 158, 145, 153, 149, 157, 147
, 155, 151, 159, 160, 168, 164, 172, 162, 170, 166, 174, 161, 169, 165, 173, 163, 171, 167
, 175, 176, 184, 180, 188, 178, 186, 182, 190, 177, 185, 181, 189, 179, 187, 183, 191, 192
, 200, 196, 204, 194, 202, 198, 206, 193, 201, 197, 205, 195, 203, 199, 207, 208, 216, 212
, 220, 210, 218, 214, 222, 209, 217, 213, 221, 211, 219, 215, 223, 224, 232, 228, 236, 226
, 234, 230, 238, 225, 233, 229, 237, 227, 235, 231, 239, 240, 248, 244, 252, 242, 250, 246
, 254, 241, 249, 245, 253, 243, 251, 247, 255};


int B[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76
, 77, 78, 79, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143
, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 16, 17, 18
, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91
, 92, 93, 94, 95, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158
, 159, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 32, 33
, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 96, 97, 98, 99, 100, 101, 102, 103, 104
, 105, 106, 107, 108, 109, 110, 111, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170
, 171, 172, 173, 174, 175, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236
, 237, 238, 239, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 112, 113, 114
, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 176, 177, 178, 179, 180
, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 240, 241, 242, 243, 244, 245, 246
, 247, 248, 249, 250, 251, 252, 253, 254, 255};


int aut[256];
for(int i=0; i<256; ++i)
    aut[i] = B[A[i]];



    Matrix<FieldElement<&F> > P(2,2);
    Matrix<FieldElement<&F> > Ph(2,2);
    Matrix<FieldElement<&F> > sigmaX(2,2);
    int s =aut[0];

    for (int a=1; a<4; ++a)
    {
        for (int b1=0; b1<F.number_of_field_elements; ++b1)
        {
            P[0][0]=field[b1]; Ph[0][0]=field_conjugated[b1];
            for (int b2=0; b2<F.number_of_field_elements; ++ b2)
            {
                P[0][1]=field[b2]; Ph[1][0]=field_conjugated[b2];
                for (int b3=0; b3<F.number_of_field_elements; ++ b3)
                {
                    P[1][0]=field[b3]; Ph[0][1]=field_conjugated[b3];
                    for (int b4=0; b4<F.number_of_field_elements; ++b4)
                    {
                        P[1][1]=field[b4]; Ph[1][1]=field_conjugated[b4];

                            for (int h=0; h<4; ++h)
                            {
                                bool is_aut = true;
                                for (int x=0; x<number_of_herm_matrices; ++x)
                                {
                                    sigmaX = Mat_herm[x];
                                    {
                                        sigmaX[0][0]=pow(sigmaX[0][0],pow(2,h));
                                        sigmaX[0][1]=pow(sigmaX[0][1],pow(2,h));
                                        sigmaX[1][0]=pow(sigmaX[1][0],pow(2,h));
                                        sigmaX[1][1]=pow(sigmaX[1][1],pow(2,h));
                                    }
                                    if (!((Ph*sigmaX*P*f_q[a]+Mat_herm[s]) == Mat_herm[aut[x]]))
                                    {
                                        is_aut =false;
                                        break;
                                    }
                                }
                                if(is_aut)
                                {
                                    cout << "a = " << f_q[a];
                                    cout << ", sigma = a |-> a^" << pow(2,h)<< endl;
                                    cout << "P = " << endl << P;
                                    cout << "S = " << endl << Mat_herm[s];
                                    cout << "---------------------------" << endl;
                                }
                            }


        }}}}


    }
return 0;
}
