#include <iostream>
#include <sstream>
#include <cmath>
#include <vector>
#include <fstream>
#include <deque>
#include <cstdlib> // for atoi

#include "finite_field.h"
#include "matrix.h"
#include "orderly.h"


extern "C" {
#include "nauty.h"
#include "cliquer.h"
}

using namespace std;



FiniteField F2(2);
FiniteField F3(3);
FiniteField F5(5);
FiniteField F7(7);

const int Poly_4[]={1,1};
FiniteField F4(2,2,Poly_4);

const int Poly_9[]={1,0};
FiniteField F9(3,2,Poly_9);

const int Poly_16[]={1,1,1,1};
FiniteField F16(2,4,Poly_16);

const int Poly_25[]={3,0};
FiniteField F25(5,2,Poly_25);

const int Poly_49[]={1,0};
FiniteField F49(7,2,Poly_49);

FiniteField F(F25);


int main(int argc, char *argv[]) // F25_2x2_herm_readstartsets
{
cout << "Anfang" << endl;
    if (argc < 3)
    {
        cout << "Usage: "<< argv[0] << " filename setsize" << endl;
        return 1;
    }
cout << "argc = " << argc << endl;
    int setsize = atoi(argv[2]);
cout << "setsize = " << setsize << endl;
    string outfilename = (string)argv[1] + "__out";
    string readfilename = argv[1];

    time_t timer0 = time(NULL);

    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }


    cout << "Körper angelegt" << endl;


/** Hermitian Matrices */

    unsigned int matrix_size=2;
    unsigned int distance=2;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (hermitian)" << endl << endl;

    int q = sqrt(F.number_of_field_elements); // F = F_{q^2}

    vector<FieldElement<&F> > field_conjugated(F.number_of_field_elements);
    vector<FieldElement<&F> > f_q;
    for (int k=0; k<F.number_of_field_elements; ++k)
    {
        field_conjugated[k] = pow(field[k],q);
        if (field[k]==field_conjugated[k])
            {f_q.push_back(field[k]); cout << field[k] << endl;}
    }

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_triangles = pow(F.number_of_field_elements,(matrix_size*(matrix_size-1))/2);
    int number_of_diagonals = pow(q,matrix_size);
    int number_of_herm_matrices = number_of_triangles * number_of_diagonals;
    vector<Matrix<FieldElement<&F> > > Mat_herm(number_of_herm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size-1))/2, 0);
        vector<vector<int> > Mat_index(number_of_triangles, M_index);

        for (int k=1; k<number_of_triangles; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size-1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_triangles; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
               // Mat_herm[k][r][r]= 0; entry is already zero
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_herm[k][r][c]= field[Mat_index[k][j]];
                    Mat_herm[k][c][r]= field_conjugated[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }
    {
        vector<int> M_index(matrix_size, 0);
        vector<vector<int> > Mat_index(number_of_diagonals, M_index);

        for (int k=1; k<number_of_diagonals; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< matrix_size-1; ++d)
            {
                if (Mat_index[k][d] == q)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }


        for (int k=1; k<number_of_diagonals; ++k)
        {
            for (int j=0; j<number_of_triangles; ++j)
            {
                Mat_herm[j+k*number_of_triangles] = Mat_herm[j];
                for (int i=0; i<matrix_size; ++i)
                    Mat_herm[j+k*number_of_triangles][i][i] = f_q[Mat_index[k][i]];
            }
        }
    }

    cout << "hermitesche Matrizen angelegt" << endl;


    graph_t *graph_herm_matrices;
    graph_herm_matrices = graph_new(number_of_herm_matrices);

    graph* g;
    int g_sz=0;
    int n = number_of_herm_matrices;
    int m =SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");

    EMPTYGRAPH(g,m,n);

    cout << "Graph angelegt; Graph enthält "<< number_of_herm_matrices << " Ecken" << endl;


    for (int k=1; k<number_of_herm_matrices; ++k)
    {
        for (int l=0; l<k; ++l)
        {
            if ((Mat_herm[k]-Mat_herm[l]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_herm_matrices,k,l);
            if (!((Mat_herm[k]-Mat_herm[l]).rank_geq(2)))
                ADDONEEDGE(g,k,l,m);
        }
    }
    cout << "Graph verknüpft; Graph enthält "<< graph_edge_count(graph_herm_matrices) << " Kanten" << endl;


    vector <Vertexset> Sk;

    ifstream fin((readfilename).c_str());
    int in;
    int counter = 0;

    Vertexset emptyset(g, graph_herm_matrices);
    Vertexset X(emptyset);

    cout << "reading startsets..." << endl;

    if (!setsize)
        Sk.push_back(emptyset);

    while((fin >> in).good())
    {
        X.add_vertex_lazy(in);
        ++counter;

        if (counter == setsize)
        {
            X.update();
            Sk.push_back(X);
            if (!(Sk.size() % 100))
                cout << Sk.size() << endl;
            X = emptyset;
            counter=0;
        }
    }
    if (!fin.eof())
    {
        cerr << "Error while reading file. Aborted." << endl;
        return 1;
    }
    if (X.setsize() != 0)
    {
        cerr << "Warning! Input file may be uncomplete!" << endl;
    }

    fin.close();

    cout << "reading complete" << endl << endl;


    time_t timer_isom0 = time(NULL);

    ofstream fout((outfilename + ".txt").c_str());

    fout << Sk[0] << endl;
    vector <Vertexset> types;
    types.push_back(Sk[0]);

    for (int ges=1; ges<Sk.size(); ++ges)
    {
        for (int t=0; t<types.size(); ++t)
        {
            if(Sk[ges].is_isom_to(types[t]))
                cout << ges + 1 << " isomorphic to type "<< t + 1 << endl;
            else
            {
                types.push_back(Sk[ges]);
                fout << Sk[ges] << endl;
            }
        }
    }

    fout.close();

    cout << outfilename + ".txt written" << endl;

    graph_free(graph_herm_matrices);
    DYNFREE(g,g_sz);
    return 0;

}
