#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cstdlib> // for atoi
#include <fstream>


#include "finite_field.h"
#include "matrix.h"

extern "C"{
#include "graph.h" // from Cliquer
}



FiniteField F2(2);
FiniteField F3(3);
FiniteField F5(5);
FiniteField F7(7);

const int Poly_4[]={1,1};
FiniteField F4(2,2,Poly_4);

const int Poly_9[]={1,0};
FiniteField F9(3,2,Poly_9);

const int Poly_16[]={1,1,1,1};
FiniteField F16(2,4,Poly_16);

const int Poly_25[]={3,0};
FiniteField F25(5,2,Poly_25);

const int Poly_49[]={1,0};
FiniteField F49(7,2,Poly_49);

const int Poly_64[]={1,1,0,0,0,0};
FiniteField F64(2,6,Poly_64);

const int Poly_81[]={2,0,0,2}; //x^4+2*x^3+2
FiniteField F81(3,4,Poly_81);

const int Poly_121[]={2,7}; // x^2+7*x+2
FiniteField F121(11,2,Poly_121);

const int Poly_169[]={2,12}; // x^2 + 12*x + 2
FiniteField F169(13,2,Poly_169);

const int Poly_256[]={1,0,1,1,1,0,0,0};// x^8 + x^4 + x^3 + x^2 + 1
FiniteField F256(2,8,Poly_256);

FiniteField F(F9);


using namespace std;

int main()
{
    time_t time_0 = time(NULL);

    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }


/** Hermitian Matrices */

    unsigned int matrix_size=4;
    unsigned int distance=4;

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (Hermitian)" << endl << endl;

    int q = sqrt(F.number_of_field_elements); // F = F_{q^2}

    vector<FieldElement<&F> > field_conjugated(F.number_of_field_elements);
    vector<FieldElement<&F> > f_q;
    for (int k=0; k<F.number_of_field_elements; ++k)
    {
        field_conjugated[k] = pow(field[k],q);
        if (field[k]==field_conjugated[k])
            {f_q.push_back(field[k]); cout << field[k] << endl;}
    }

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_triangles = pow(F.number_of_field_elements,(matrix_size*(matrix_size-1))/2);
    int number_of_diagonals = pow(q,matrix_size);
    int number_of_herm_matrices = number_of_triangles * number_of_diagonals;
    vector<Matrix<FieldElement<&F> > > Mat_herm(number_of_herm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size-1))/2, 0);
        vector<vector<int> > Mat_index(number_of_triangles, M_index);

        for (int k=1; k<number_of_triangles; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size-1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_triangles; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
               // Mat_herm[k][r][r]= 0; entry is already zero
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_herm[k][r][c]= field[Mat_index[k][j]];
                    Mat_herm[k][c][r]= field_conjugated[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }
    {
        vector<int> M_index(matrix_size, 0);
        vector<vector<int> > Mat_index(number_of_diagonals, M_index);

        for (int k=1; k<number_of_diagonals; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< matrix_size-1; ++d)
            {
                if (Mat_index[k][d] == q)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }


        for (int k=1; k<number_of_diagonals; ++k)
        {
            for (int j=0; j<number_of_triangles; ++j)
            {
                Mat_herm[j+k*number_of_triangles] = Mat_herm[j];
                for (int i=0; i<matrix_size; ++i)
                    Mat_herm[j+k*number_of_triangles][i][i] = f_q[Mat_index[k][i]];
            }
        }
    }

    cout << "hermitesche Matrizen angelegt" << endl;




    int number = 1; // number of start vertices
    int start_set[] = {0};

    cout << "start set: ";
    for (int i=0; i<number; ++i) cout << start_set[i] << " ";
    cout << endl;

    vector<int> vertices_adjacent_to_start_set;
    for (int i=0; i<number_of_herm_matrices; ++i)
    {
        bool adjacent = true;
        for (int j=0; j<number; ++j)
        {
            adjacent = adjacent && (Mat_herm[i]-Mat_herm[start_set[j]]).rank_geq(distance);
        }
        if (adjacent)
            vertices_adjacent_to_start_set.push_back(i);
    }

    int number_of_adjacent_vertices = vertices_adjacent_to_start_set.size();

    graph_t *graph_of_adjacent_vertices;
    graph_of_adjacent_vertices = graph_new(number_of_adjacent_vertices);

    cout << "Graph_of_adjacent_vertices erstellt; Graph enthält "<< number_of_adjacent_vertices << " Ecken" << endl;

    for (int i=1; i<number_of_adjacent_vertices; ++i)
    {
        for (int j=0; j<i; ++j)
        {
            if ((Mat_herm[vertices_adjacent_to_start_set[i]]-Mat_herm[vertices_adjacent_to_start_set[j]]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_of_adjacent_vertices,i,j);
        }
    }

    cout << "Graph_of_adjacent_vertices verknüpft" << endl;


    // choose graph
    graph_t *g = graph_of_adjacent_vertices;
    int n = number_of_adjacent_vertices;
    ASSERT(graph_test(g,NULL));

    graph_write_dimacs_binary_file(g,NULL,"graph.dimacsbin");

    ofstream fout("adjacent.txt");
    for(int i=0;i<number_of_adjacent_vertices;++i)
        fout << vertices_adjacent_to_start_set[i] << " ";
    fout << endl;
    fout.close();

    return 0;
}
