#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cstdlib> // for atoi
#include <fstream>


#include "finite_field.h"
#include "matrix.h"

extern "C"{
#include "graph.h" // from Cliquer
}



FiniteField F2(2);
FiniteField F3(3);
FiniteField F5(5);
FiniteField F7(7);

const int Poly_4[]={1,1};
FiniteField F4(2,2,Poly_4);

const int Poly_9[]={1,0};
FiniteField F9(3,2,Poly_9);

const int Poly_16[]={1,1,1,1};
FiniteField F16(2,4,Poly_16);

const int Poly_25[]={3,0};
FiniteField F25(5,2,Poly_25);

const int Poly_49[]={1,0};
FiniteField F49(7,2,Poly_49);

const int Poly_64[]={1,1,0,0,0,0};
FiniteField F64(2,6,Poly_64);

FiniteField F(F2);


using namespace std;

int main()
{
    time_t time_0 = time(NULL);

    vector<FieldElement<&F> > field(F.number_of_field_elements);
    for (int k=1; k<F.number_of_field_elements; ++k)
    {
        field[k] = field[k-1];

        ++field[k][0];

        for (int d=0; d< (const signed int) F.deg-1; ++d)
        {
            if (field[k][d] == (const signed int) F.p)
            {
                field[k][d] = 0;
                ++field[k][d+1];
            }
        }
    }


/** Symmetric Matrices */

    unsigned int matrix_size=5;
    unsigned int distance=4;

    if(distance > matrix_size)
    {
        cout << "error: distance > matrix_size " << endl;
        return 1;
    }

    cout << "F" << F.number_of_field_elements << ", " << matrix_size << "x" << matrix_size << " (symmetric)" << endl << endl;

    Matrix<FieldElement<&F> > M(matrix_size, matrix_size);

    int number_of_symm_matrices = pow(F.number_of_field_elements,(matrix_size*(matrix_size+1))/2);
    vector<Matrix<FieldElement<&F> > > Mat_symm(number_of_symm_matrices, M);

    {
        vector<int> M_index((matrix_size*(matrix_size+1))/2, 0);
        vector<vector<int> > Mat_index(number_of_symm_matrices, M_index);

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            Mat_index[k] = Mat_index[k-1];

            ++Mat_index[k][0];

            for (int d=0; d< (const signed int) (matrix_size*(matrix_size+1))/2-1; ++d)
            {
                if (Mat_index[k][d] == (const signed int) F.number_of_field_elements)
                {
                    Mat_index[k][d] = 0;
                    ++Mat_index[k][d+1];
                }
            }
        }

        for (int k=1; k<number_of_symm_matrices; ++k)
        {
            int j=0;
            for (unsigned int r=0; r<matrix_size; ++r)
            {
                Mat_symm[k][r][r]= field[Mat_index[k][j]];
                ++j;
                for (unsigned int c=0; c<r; ++c)
                {
                    Mat_symm[k][r][c]= field[Mat_index[k][j]];
                    Mat_symm[k][c][r]= field[Mat_index[k][j]];
                    ++j;
                }
            }
        }
    }

    cout << "symmetrische Matrizen angelegt" << endl;



    int number = 1; // number of start vertices
    int start_set[] = {0};

    cout << "start set: ";
    for (int i=0; i<number; ++i) cout << start_set[i] << " ";
    cout << endl;

    vector<int> vertices_adjacent_to_start_set;
    for (int i=0; i<number_of_symm_matrices; ++i)
    {
        bool adjacent = true;
        for (int j=0; j<number; ++j)
        {
            adjacent = adjacent && (Mat_symm[i]-Mat_symm[start_set[j]]).rank_geq(distance);
        }
        if (adjacent)
            vertices_adjacent_to_start_set.push_back(i);
    }

    int number_of_adjacent_vertices = vertices_adjacent_to_start_set.size();

    graph_t *graph_of_adjacent_vertices;
    graph_of_adjacent_vertices = graph_new(number_of_adjacent_vertices);

    cout << "Graph_of_adjacent_vertices erstellt; Graph enthält "<< number_of_adjacent_vertices << " Ecken" << endl;

    for (int i=1; i<number_of_adjacent_vertices; ++i)
    {
        for (int j=0; j<i; ++j)
        {
            if ((Mat_symm[vertices_adjacent_to_start_set[i]]-Mat_symm[vertices_adjacent_to_start_set[j]]).rank_geq(distance))
                GRAPH_ADD_EDGE(graph_of_adjacent_vertices,i,j);
        }
    }

    cout << "Graph_of_adjacent_vertices verknüpft" << endl;


    // choose graph
    graph_t *g = graph_of_adjacent_vertices;
    int n = number_of_adjacent_vertices;
    ASSERT(graph_test(g,NULL));

    graph_write_dimacs_binary_file(g,NULL,"graph.dimacsbin");

    ofstream fout("adjacent.txt");
    for(int i=0;i<number_of_adjacent_vertices;++i)
        fout << vertices_adjacent_to_start_set[i] << " ";
    fout << endl;
    fout.close();

    return 0;
}
