

#ifndef MATRIX_TEMPLATE_CLASS_DEFINED
#define MATRIX_TEMPLATE_CLASS_DEFINED


#include <iomanip>
#include <iostream>
#include <cassert>

using namespace std;

template <typename Type>
class Matrix;

template <typename Type>
ostream & operator<< (ostream & os, const Matrix<Type> & mat);

template <typename Type>
class Matrix
{
    protected:
        Type* ptr;
        unsigned int r,c;

    protected:
        bool allocate(unsigned int n);
        void de_allocate();

    public:
        Matrix(unsigned int rows, unsigned int columns, const Type &entry);


        Matrix(unsigned int rows, unsigned int columns, Type* adress= (Type*) 0 );

        template <unsigned int R, unsigned int C>
        Matrix(Type m[R][C]);

        Matrix(const Matrix & mat);

        Matrix & operator=(const Matrix & mat);

        Matrix & operator+=(const Matrix & M);
        Matrix operator+(const Matrix & M) const;
        Matrix & operator-=(const Matrix & M);
        Matrix operator-(const Matrix & M) const;
        Matrix & operator*=(const Matrix & M);
        Matrix operator*(const Matrix & M) const;
        Matrix & operator*=(const Type & lambda);
        Matrix operator*(const Type & lambda) const;
        Type* operator[](int k);
        Type* operator[](int k) const;
        bool operator==(const Matrix & M) const;

        bool rank_geq(unsigned int min_rank);

        void print_to_latex(ostream & os=cout) const;

        ~Matrix();

    friend ostream & operator<< <Type>(ostream & os, const Matrix & mat);
};



template <typename Type>
bool Matrix<Type>::allocate(unsigned int n)
{
    ptr = new (nothrow) Type[n];

    return (ptr != (Type*)0);
}

template <typename Type>
void Matrix<Type>::de_allocate()
{
    delete[] ptr;
}

template <typename Type>
Matrix<Type>::Matrix(unsigned int rows, unsigned int columns, const Type &entry): r(rows), c(columns)
{
    unsigned int n=r*c;
    if (n==0)
    {
        ptr = (Type*) 0;
    }
    else
    {
        if(!(allocate(n)))
        {
            cerr << "error in function Matrix<Type>::Matrix(unsigned int rows, unsigned int columns, const Type &entry):" << endl
                 << "It was not possible to allocate enough memory. Matrix will have size 0x0!" << endl;
            r = c = 0;
        }
        else
        {
            for(unsigned int j=0; j<n; ++j)
                ptr[j] = entry;
        }
    }
}

template <typename Type>
Matrix<Type>::Matrix(unsigned int rows, unsigned int columns, Type* adress): r(rows), c(columns)
{
    unsigned int n=r*c;
    if (n==0)
    {
        ptr = (Type*) 0;
    }
    else
    {
        if(!(allocate(n)))
        {
            cerr << "error in function Matrix<Type>::Matrix(unsigned int rows, unsigned int columns, Type* adress):" << endl
                 << "It was not possible to allocate enough memory. Matrix will have size 0x0!" << endl;
            r = c = 0;
        }
        else if(adress != (Type*) 0)
        {
            for(unsigned int j=0; j<n; ++j)
                ptr[j] = adress[j];
        }
    }
}

template <typename Type>
template <unsigned int R,unsigned int C>
Matrix<Type>::Matrix(Type m[R][C]): r(R), c(C)
{
    unsigned int n=r*c;
    if (n==0)
    {
        ptr = (Type*) 0;
    }
    else
    {
        if(!(allocate(n)))
        {
            cerr << "error in function Matrix<Type>::Matrix(Type[R][C] m):" << endl
                 << "It was not possible to allocate enough memory. Matrix will have size 0x0!" << endl;
            r = c = 0;
        }
        else
        {
            for (unsigned int r_=0; r_<R; ++r_)
            {
                for (unsigned int c_=0; c_<C; ++c_)
                {
                    ptr[c_+r_*C] = m[r_][c_];
                }
            }
        }
    }
}

template <typename Type>
Matrix<Type>::Matrix(const Matrix & mat): r(mat.r), c(mat.c)
{
    unsigned int n=r*c;
    if (n==0)
    {
        ptr = (Type*) 0;
    }
    else
    {
        if(!(allocate(n)))
        {
            cerr << "error in function Matrix<Type>::Matrix(const Matrix & mat):" << endl
                 << "It was not possible to allocate enough memory. Matrix will have size 0x0!" << endl;
            r = c = 0;
        }
        else
        {
            for(unsigned int j=0; j<n; ++j)
                ptr[j] = mat.ptr[j];
        }
    }
}


template <typename Type>
Matrix<Type> & Matrix<Type>::operator=(const Matrix & mat)
{
    if (this != &mat)
    {
        if (r*c != mat.r*mat.c)
        {
            r = mat.r;
            c = mat.c;

            de_allocate();
            if(!(allocate(r*c)))
            {
                cerr << "error in function Matrix & Matrix<Type>::operator=(const Martix & mat):" << endl
                     << "It was not possible to allocate enough memory. Matrix will have size 0x0!" << endl;
                r = c = 0;
            }
            else
            {
                for(unsigned int j=0; j<r*c; ++j)
                    ptr[j] = mat.ptr[j];
            }
        }
        else
        {
            r = mat.r;
            c = mat.c;
            for(unsigned int j=0; j<r*c; ++j)
                ptr[j] = mat.ptr[j];
        }
    }

    return *this;
}



template <typename Type>
Matrix<Type> & Matrix<Type>::operator+=(const Matrix & M)
{
    assert((this->c==M.c) && (this->r==M.r));
    for (unsigned int j=0; j<this->c*this->r; ++j)
        this->ptr[j] += M.ptr[j];
    return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator+(const Matrix & M) const
{
    Matrix<Type> M2(*this);
    return (M2 += M);
}

template <typename Type>
Matrix<Type> & Matrix<Type>::operator-=(const Matrix & M)
{
    assert((this->c==M.c) && (this->r==M.r));
    for (unsigned int j=0; j<this->c*this->r; ++j)
        this->ptr[j] -= M.ptr[j];
    return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator-(const Matrix & M) const
{
    Matrix<Type> M2(*this);
    return (M2 -= M);
}

template <typename Type>
Matrix<Type> & Matrix<Type>::operator*=(const Matrix & M)
{
    *this = *this * M;
    return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator*(const Matrix & M) const
{
    assert (this->c == M.r);
    Matrix<Type> Erg(this->r,M.c);
    for (int i=0; i<this->r; ++i)
    {
        for (int j=0; j<M.c; ++j)
        {
            Erg[i][j]=0;
            for (int k=0; k<this->c; ++k)
                Erg[i][j] += (*this)[i][k]*M[k][j];
        }
    }
    return Erg;
}

template <typename Type>
Matrix<Type> & Matrix<Type>::operator*=(const Type & lambda)
{
    for (unsigned int j=0; j<this->c*this->r; ++j)
        this->ptr[j] *= lambda;
    return *this;
}

template <typename Type>
Matrix<Type> Matrix<Type>::operator*(const Type & lambda) const
{
    Matrix<Type> M(*this);
    return (M *= lambda);
}

template <typename Type>
Type* Matrix<Type>::operator[](int k)
{
    return &(this->ptr[k*this->c]);
}

template <typename Type>
Type* Matrix<Type>::operator[](int k) const
{
    return &(this->ptr[k*this->c]);
}

template <typename Type>
bool Matrix<Type>::operator==(const Matrix & M) const
{
    if (this == &M)
        return true;

    if (r != M.r || c != M.c)
        return false;

    int s = r*c;
    for (int i=0; i<s; ++i)
    {
        if (ptr[i] != M.ptr[i])
            return false;
    }

    return true;
}

template <typename Type>
bool Matrix<Type>::rank_geq(unsigned int min_rank)
{
    if (min_rank > r)
        return false;
    if (min_rank > c)
        return false;

    // copy matrix
    Type** M = new Type* [r];
    for (unsigned int i=0; i<r; ++i)
    {
        M[i] = new Type [c];
        for (unsigned int j=0; j<c; ++j)
            M[i][j] = this->operator[](i)[j];
    }

    unsigned int rk = 0;
    unsigned int pivot;
    unsigned int k;
    Type* tmp;

    for (unsigned int j=0; j<c; ++j)
    {
        // look for pivot element in j-th column
        pivot = -1;
        k = rk;
        while (pivot == -1 && k<r)
        {
            if (M[k][j]!=0)
                pivot = k;
            ++k;
        }

        // if nonzero element is found swap and subtract rows and increase rank
        if (pivot != -1)
        {
            // swap
            tmp = M[pivot];
            M[pivot] = M[rk];
            M[rk] = tmp;

            // subtract
            for (unsigned int ii=rk+1; ii<r; ++ii)
            {
                for (unsigned int jj=j+1; jj<c; ++jj)
                    M[ii][jj] -= (M[ii][j]/M[rk][j]) * M[rk][jj];
                M[ii][j] = 0;
            }

            // increase rank
            ++rk;
        }

        // if already min_rank linear independent rows are found, return true
        if(rk == min_rank)
        {
            for (unsigned int i=0; i<r; ++i)
            {
                delete[] M[i];
            }
            delete[] M;
            return true;
        }
    }

    // return false, because rk is still smaller than min_rank
    for (unsigned int i=0; i<r; ++i)
    {
        delete[] M[i];
    }
    delete[] M;
    return false;
}

template <typename Type>
void Matrix<Type>:: print_to_latex(ostream & os) const
{
    if (c*r==0)
    {
        os << "(empty)" << endl;
    }
    else
    {
        os << "$\\begin{pmatrix}" << endl;
        for (unsigned int r_=0; r_<r; ++r_)
        {
            os << ptr[r_*c];
            for (unsigned int c_=1; c_<c; ++c_)
            {
                os << " & " << ptr[c_+r_*c];
            }
            if (r_<r-1)
                os << " \\\\";
            os << endl;
        }
        os << "\\end{pmatrix}$" << endl;
    }
    return;
}


template <typename Type>
Matrix<Type>::~Matrix()
{
    de_allocate();
}


template <typename Type>
ostream & operator<< (ostream & os, const Matrix<Type> & mat)
{
    if (mat.c*mat.r==0)
    {
        os << "(empty)" << endl;
    }
    else
    {
        for (unsigned int r_=0; r_<mat.r; ++r_)
        {
            for (unsigned int c_=0; c_<mat.c; ++c_)
            {
                os << setw(6) << mat.ptr[c_+r_*mat.c] << " ";
            }
            os << endl;
        }
    }

    return os;
}


#endif  // MATRIX_TEMPLATE_CLASS_DEFINED

