#ifndef ORDERLY_DEFINED
#define ORDERLY_DEFINED

#include <iostream>
#include <vector>
#include <cmath>


extern "C" {
#include "nauty.h"
#include "naututil.h"
#include "cliquer.h"
}

class Vertexset;

std::ostream & operator<< (std::ostream & os, const Vertexset & X);

class Vertexset
{
    private:
        int* lab;
        int* ptn;
        int n;
        int sz;
        int* orbits;

    public:
        graph* g_dist1;
        graph_t* g_min_dist;

    public:
        Vertexset();
        Vertexset(graph* const graph_dist1, graph_t* graph_min_dist);
        Vertexset(const Vertexset &X);
        Vertexset operator=(const Vertexset &X);
        Vertexset add_vertex(int vertex);
        Vertexset add_vertex_lazy(int vertex);
        Vertexset update();
        int setsize() const;
        int graph_size() const;
        int get_lab(int i) const;
        int get_orbits(int i) const;
        bool is_isom_to(const Vertexset &X) const;
        double write_automs(std::string filename, bool writemarkers, bool append);
        ~Vertexset();

    friend std::ostream & operator<< (std::ostream & os, const Vertexset & X);
};


std::vector<Vertexset> augment_vertexset(Vertexset X);

#endif // ORDERLY_DEFINED
